package mapper;

import com.cycling74.max.*;
import java.util.*;
import java.awt.*;

public class BoxRegion
{
	private static final int _rwidth = 120;
	private static final int _rheight = 80;
	
	private HashMap<Point, LinkedList<MaxBox>> _regionMap;
	
	public BoxRegion()
	{
		_regionMap = new HashMap<Point, LinkedList<MaxBox>>();
	}
	
	public void RemoveAll()
	{
		_regionMap.clear();
	}
	
	public void RemoveAllBoxes() {
		Collection<LinkedList<MaxBox>> allboxes = _regionMap.values();
		Iterator<LinkedList<MaxBox>> boxes = allboxes.iterator();
		while (boxes.hasNext()) {
			Iterator<MaxBox> box = boxes.next().iterator();
			while (box.hasNext()) {
				MaxBox nextbox = box.next();
				if (nextbox != null) {
					try {
						MaxPatcher parent = nextbox.getPatcher();
						MaxBox thebox = parent.getNamedBox(nextbox.getName());
						if (thebox != null)
							thebox.remove();
					} catch (Exception e) {
						System.out.println("caught exception removeall");
					}
				}
			}
		}
		_regionMap.clear();
	}
	
	//return box at location x,y; return null if no box located
	public MaxBox GetSelection( int x, int y )
	{
		//mouse pos to region
		LinkedList<Point> mindx = GetRegionIndexs(new int[]{x, y, x, y});
		
		if( mindx.size() == 0 )
		{
			return null;
		}
		
		if( ! _regionMap.containsKey(mindx.getFirst() ))
		{
			return null;
		}
		Iterator<MaxBox> boxiter = _regionMap.get(mindx.getFirst()).iterator();
		
		//check if mouse is contained in any of the boxes in the region
		while( boxiter.hasNext() )
		{
			MaxBox curbox = boxiter.next();
			
			int[] boxloc = curbox.getRect();
			if( ( x <= boxloc[2] ) && ( x >= boxloc[0]) && ( y <= boxloc[3]) && ( y >= boxloc[1]) )
			{
				return curbox;
			}
		}
		
		return null;
	}
	
	public LinkedList<MaxBox> GetGroupSelection( int[] rect )
	{
		LinkedList<MaxBox> selections = new LinkedList<MaxBox>();
		LinkedList<Point> regs = GetRegionIndexs(rect);
		
		Iterator<Point> inditer = regs.iterator();
		//iterate through all regions in the group selection rectangle
		while( inditer.hasNext() ) 
		{
			Point next = inditer.next();
			if( _regionMap.containsKey(next) )
			{
				Iterator<MaxBox> boxiter = _regionMap.get(next).iterator();
				//iterate through all boxes in this region
				while( boxiter.hasNext() )
				{
					MaxBox nextbox = boxiter.next();
					int[] boxloc = nextbox.getRect();
					
					//if upper left corner is contained in selection
					if( ( boxloc[0] <= rect[2] ) && ( boxloc[0] >= rect[0]) 
							&& ( boxloc[1] <= rect[3]) && ( boxloc[1] >= rect[1]) )
					{
						if(! selections.contains(nextbox) )
							selections.add(nextbox);
					}
					//if upper right corner is contained in selection
					else if ( ( boxloc[2] <= rect[2] ) && ( boxloc[2] >= rect[0]) 
							&& ( boxloc[1] <= rect[3]) && ( boxloc[1] >= rect[1]) )
					{
						if(! selections.contains(nextbox) )
							selections.add(nextbox);
					}
					//if lower right corner is contained in selection
					else if ( ( boxloc[2] <= rect[2] ) && ( boxloc[2] >= rect[0]) 
							&& ( boxloc[3] <= rect[3]) && ( boxloc[3] >= rect[1]) )
					{
						if(! selections.contains(nextbox) )
							selections.add(nextbox);
					}
					//if lower left corner is contained in selection
					else if( ( boxloc[0] <= rect[2] ) && ( boxloc[0] >= rect[0]) 
							&& ( boxloc[3] <= rect[3]) && ( boxloc[3] >= rect[1]) )
					{
						if(! selections.contains(nextbox) )
							selections.add(nextbox);
					}
					//TODO if bottom/right middle section is contained in selection
					else if( ((( rect[1] < boxloc[3] ) && ( rect[1] > boxloc[1]))
							|| (( boxloc[1] > rect[1] ) && ( boxloc[1] < rect[3] )))
							&& ( rect[0] > boxloc[0]) && ( rect[0] < boxloc[2] ) )
					{
						if(! selections.contains(nextbox) )
							selections.add(nextbox);
					}
					//if upper/left middle section is contained in selection
					else if( ( rect[3] < boxloc[3] ) && ( rect[3] > boxloc[1]) 
							&& ( (( rect[2] > boxloc[0]) && ( rect[2] < boxloc[2] ))
							|| ((boxloc[0] > rect[0]) && ( boxloc[0] < rect[2])) ) )
					{
						if(! selections.contains(nextbox) )
							selections.add(nextbox);
					}
				}
			}
		}
		return selections;
	}
	
	//add a new box to all regions it is in
	public void AddBox( MaxBox addbox )
	{
		int[] loc = addbox.getRect();
		LinkedList<Point> regs = GetRegionIndexs(loc);
		
		if( regs.size() == 0 )
		{
			return;
		}
		
		Iterator<Point> iter = regs.iterator();
		while( iter.hasNext() )
		{
			Point next = iter.next();
			if( _regionMap.containsKey(next) )
			{
				_regionMap.get(next).add(addbox);
			}
			else
			{
				LinkedList<MaxBox> newlist = new LinkedList<MaxBox>();
				newlist.add(addbox);
				_regionMap.put(next, newlist);
			}
		}
		
	}
	
	//remove a box from all the regions it is a member of
	public boolean RemoveBox( MaxBox rembox )
	{
		boolean removed = false;
		int[] loc = rembox.getRect();
		if( loc == null )
		{
			return false;
		}
		
		LinkedList<Point> regs = GetRegionIndexs(loc);
		
		if( regs.size() == 0 )
		{
			return false;
		}
		
		Iterator<Point> iter = regs.iterator();
		while( iter.hasNext() )
		{
			Point next = iter.next();
			if( _regionMap.containsKey(next) )
			{
				if( _regionMap.get(next).remove(rembox) )
				{
					removed = true;
				}
				else
				{
					System.out.println("box not found in region: (" + next.x + ", " + next.y + ")");
				}
			}
		}
		
		return removed;
	}
	
	//compute all region indexs a given rectangle is in
	private LinkedList<Point> GetRegionIndexs(int[] rec)
	{	
		LinkedList<Point> indexs = new LinkedList<Point>();
		Point ind1 = new Point(0,0);
		ind1.x = rec[0]/_rwidth;
		ind1.y = rec[1]/_rheight;
		
		Point ind2 = new Point(0,0);
		ind2.x = rec[2]/_rwidth;
		ind2.y = rec[3]/_rheight;
		
		if( ( ind2.x - ind1.x < 0 ) || ( ind2.y - ind1.y < 0 ) || ( ind1.x < 0 )
				|| ( ind1.y < 0 ) || (ind2.x < 0 ) || (ind2.y < 0 ))
		{
			return indexs;
		}
		
		for( int i = ind1.x; i <= ind2.x; i++ )
		{
			for(int j = ind1.y; j <= ind2.y; j++ )
			{
				Point nextindx = new Point(i,j);
				indexs.add(nextindx);
			}
		}
		
		
		return indexs;
	}
	
}
