package mapper;

import com.cycling74.max.MaxBox;

public class CcMapControl extends MapControl {
	public static final int ABSOLUTE = 0;
	public static final int BINOFF = 1;
	public static final int TWOS = 2;
	public static final int SIGNBIT1 = 3;
	public static final int SIGNBIT2 = 4;
	private int _encoder;
	private float _inval;
	
	public CcMapControl(MaxBox thebox, String name, int type, float min,
			float max) {
		super(thebox, name, type, min, max);
		_encoder = ABSOLUTE;
	}

	public CcMapControl(MaxBox thebox, String name, int type, float min,
			float max, float percision, float curve, boolean sendupoff,
			boolean sendvelocity, boolean banger) {
		super(thebox, name, type, min, max, percision, curve, sendupoff, sendvelocity,
				banger);
		_encoder = ABSOLUTE;
	}
	
	public CcMapControl(MaxBox thebox, String name, int type, float min,
			float max, float percision, float curve, int encoder, boolean sendupoff,
			boolean sendvelocity, boolean banger) {
		super(thebox, name, type, min, max, percision, curve, sendupoff, sendvelocity,
				banger);
		this.encoder(encoder);
	}
	
	public void send(int val) {
		//if (_type != CONTROLLER){
		switch( _encoder) {
		case ABSOLUTE:
			if (val == 0) {
				if (_sendupoff) {
					if (_banger)
						_theBox.bang();
					else {
						_curval = _minval;
						_theBox.send(_curval);
					}
				}
			}
			else {
				float newval;
				if (_banger)
					_theBox.bang();
				else {
					if ((_precision > 0) && (val != 127)) {
						newval = Math.round(_precisionmult * compute(val)) * _precision;
						if (newval != _curval) {
							_curval = newval;
							_theBox.send(_curval);
						}
					}
					else {
						_curval = compute(val);
						_theBox.send(_curval);
					}
				}
			}
			return;
		case BINOFF:
			val = val - 64;
			break;
		case TWOS:
			if (val > 64) 
				val = val - 128;
			break;
		case SIGNBIT1:
			if (val < 64)
				val = -val;
			else
				val = val - 64;
			break;
		case SIGNBIT2:
			if (val > 64)
				val = 64 - val;
			break;
		}
		_inval = _inval + val;

		if (_minval != _maxval) {
			if (_precision > 0) { //increment/decrement by precision
				_curval = Math.round(_curval*_precisionmult + val )*_precision;
			}
			else { //act like absolute type
				if (_inval > 127)
					if(_sendupoff)
						_inval = 0;
					else
						_inval = 127;
				else if (_inval < 0)
					if(_sendupoff)
						_inval = 127;
					else
						_inval = 0;
				_curval = compute(_inval);
			}
			//clip/rollover for max/min values
			if (_maxval > _minval) {
				if (_curval > _maxval)
					if (_sendupoff) //rollover
						_curval = _minval;
					else //clip
						_curval = _maxval;
				else if (_curval < _minval)
					if (_sendupoff) //rollover
						_curval = _maxval;
					else //clip
						_curval = _minval;
			}
			else { //max < min
				if (_curval < _maxval)
					if (_sendupoff) //rollover
						_curval = _minval;
					else //clip
						_curval = _maxval;
				else if (_curval > _minval)
					if (_sendupoff) //rollover
						_curval = _minval;
					else //clip
						_curval = _maxval;
			}
		}
		//send encoder delta value for minval == maxval && precision == 0
		else if (_precision == 0) {
			_curval = val;
		}
		//no min, max for this control
		else
			_curval = Math.round(_curval*_precisionmult + val )*_precision;
		
		_theBox.send(_curval);
	}
	
	public int encoder() {
		return _encoder;
	}
	
	public void encoder(int enc) {
		switch(enc) {
		case BINOFF:
			break;
		case TWOS:
			break;
		case SIGNBIT2:
			break;
		case SIGNBIT1:
			break;
			default:
				enc = ABSOLUTE;
		}
		_encoder = enc;
	}
}
