package mapper;


import com.cycling74.max.*;
//import java.util.*;

public class MapControl {
	public static final int KEY = 0;
	public static final int NOTE = 1;
	public static final int CONTROLLER = 2;
	public static final int CC = 3;
	
	protected MaxBox _theBox;
	protected String _boxName;
	protected int _type;
	protected float _curval;
	protected float _minval;
	protected float _maxval;
	//protected float _slope;
	protected boolean _sendupoff;
	protected boolean _sendvelocity;
	protected boolean _banger;
	protected float _precision;
	protected float _precisionmult;
	protected float _curve;
	protected float _c0;
	protected float _c1;
	protected float _c2;
	protected float _dscale = (float)(1.0/127.0);
	protected float _rscale = _maxval - _minval;
	protected float _base;
	
	public MapControl (MaxBox thebox, String name, int type, float min, float max) {
		this(thebox, name, type, min, max, 0, 0, true, true, false);		
	}
	public MapControl (MaxBox thebox, String name, int type, float min, float max, 
			float percision, float curve, boolean sendupoff, boolean sendvelocity, boolean banger) {
		_theBox = thebox;
		_boxName = name;
		this.type(type);
		_minval = min;
		_maxval = max;
		_curval = min;
		_sendupoff = sendupoff;
		_sendvelocity = sendvelocity;
		_banger = banger;
		_curve = curve;
		precision(percision);
		computeConstants();
	}
	
	private void computeConstants() {
		_rscale = _maxval - _minval;
		if (_curve == 0) {
			_c0 = _rscale*_dscale;
			_c1 = _minval;
		}
		else {
			_base = Math.abs(_curve)+1;
			if (_curve > 0) {
				_c0 = (float)(_rscale/Math.log(_base));
				_c1 = (_base - 1)*_dscale;
				_c2 = 1;
			}
			else {				
				_c0 = _rscale/(_base-1);
				_c1 = _minval - _rscale/(_base - 1); 				
			}
		}
	}
	
	public String name() {
		return _boxName;
	}
	
	public void name(String name) {
		_boxName = name;
	}
	
	public float compute(float val) {
		if (_curve == 0) {
			return _c0 * val + _c1;
		}
		else if (_curve > 0) {
			return (float)(_c0*Math.log(_c1*val+_c2) + _minval);
		}
		else {
			return (float)(_c0*Math.pow(_base, val*_dscale)+_c1);
		}
	}
	
	public void send(int val) {
		//if (_type != CONTROLLER){
		if (val == 0) {
			if (_sendupoff) {
				if (_banger)
					_theBox.bang();
				else {
					_curval = _minval;
					_theBox.send(_curval);
				}
			}
		}
		else {
			float newval;
			if (_banger)
				_theBox.bang();
			else {
				if (_sendvelocity) {
					newval = compute(val);
					if ((_precision > 0) && (val != 127)) {
						newval = Math.round(_precisionmult * newval) * _precision;
						if (newval != _curval) {
							_curval = newval;
							_theBox.send(_curval);
						}
					}
					else {
						_curval = newval;
						_theBox.send(_curval);
					}
				}
				else {
					//toggle min/max if !(sendupoff && sendvelocity)
					if ((_type == NOTE) && (! _sendupoff)) { 
						if (_curval == _maxval){
							_curval = _minval;
						}
						else {
							_curval = _maxval;
						}
					}
					else
						_curval = _maxval;
					_theBox.send(_curval);
				}
				
			}
		}
	}
	
	public int type() {
		return _type;
	}
	
	public void type(int t) {
		switch (t) {
		case KEY:
			break;
		case NOTE:
			break;
		case CC:
			break;
			default:
				t = CONTROLLER;
		}
		_type = t;
	}
	
	public float min() {
		return _minval;
	}
	
	public void min(float m) {
		_minval = m;
		//_slope = (_maxval-_minval)/(float)127.0;
		computeConstants();
	}
	
	public float max() {
		return _maxval;
	}
	
	public void max(float m) {
		_maxval = m;
		//_slope = (_maxval-_minval)/(float)127.0;
		computeConstants();
	}
	
	public void precision(float p) {
		if (p > 0) {
			_precision = p;
			_precisionmult = (float)1.0/p;
			_curval = Math.round(_precisionmult*_curval)*_precision;
		}
		else {
			_precision = 0;
		}
			
	}
	
	public float precision() {
		return (float)_precision;
	}
	
	public void sendUpOff(boolean n) {
		_sendupoff = n;
	}
	
	public boolean sendUpOff() {
		return _sendupoff;
	}
	
	public boolean sendVelocity() {
		return _sendvelocity;
	}
	
	public void sendVelocity(boolean v) {
		_sendvelocity = v;
	}
	
	public void banger(boolean b) {
		_banger = b;
	}
	
	public boolean banger() {
		return _banger;
	}
	
	public MaxBox theBox() {
		return _theBox;
	}
	
	public void theBox(MaxBox thebox) {
		_theBox = thebox;
	}
	
	public float curve() {
		return _curve;
	}
	
	public void curve(float c) {
		if (_curve != c) {
			_curve = c;
			computeConstants();
		}
	}
}
