package mapper;

import com.cycling74.max.*;
import java.util.*;

public class MaxMapper extends MaxObject {
	
	private MaxBox _myBox; //mxj box
	private MaxBox _guiBox; //interface bpatcher
	private MaxBox _highlightedBox; //highlight of box hovered over
	private MaxPatcher _myParent; //parent patcher
	private MaxWindow _myWindow; //patcher window
	private BoxRegion _boxes; //all boxes that can be mapped in the patch
	private BoxRegion _mapMarkers; //markers (panels) indicating mapped boxes
	private MaxBox _lastMarker; //last marker hovered over
	private MaxBox _selectedMarker; //marker 
	private boolean _learning;
	private HashMap<Integer, MapControl> _mapping; //map from midi/key code to controlled box
	private HashMap<String, Integer> _invMapping; //map from name of controlled box to midi/key code
	private MaxBox _highlighter; //highlighter for hovered over box
	private int _lastHash; //hash code of last midi/key
	private boolean _mousedown; //true if mouse button is down
	private boolean _chasing;
	private LinkedList<String> _selectedBoxes; //names of boxes selected to be mapped
	private HashMap<String,MaxBox> _highlightMap; //highlights of boxes selected to be mapped
	private MapControl _lastDisplayed; //last displayed control
	private boolean _blockRepeatMap; //block repeated mapping of one midi/key when multiple
	private int[] _mouseLocation;
	private Set<Character> _devices; //midi devices enabled
	private int _queueHash; //next hash code whose mapping configuration is to be displayed
	private MaxClock _delayDisplay; //display mapping configuration after small delay
	
	private static final String[] INLET_ASSIST = new String[]{
		"control in"
	};
	private static final String[] OUTLET_ASSIST = new String[]{
		"to gui",
		"to coll",
		"to thispatcher"
	};
	
	public MaxMapper(Atom[] args) {
		declareInlets(new int[]{DataTypes.ALL});
		declareOutlets(new int[]{DataTypes.ALL, DataTypes.ALL, DataTypes.ALL});
		
		setInletAssist(INLET_ASSIST);
		setOutletAssist(OUTLET_ASSIST);
		
		_myBox = getMaxBox();
		_myParent = getParentPatcher();
		_myWindow = _myParent.getWindow();
		_guiBox = _myParent.getNamedBox("maxmappergui");
		_boxes = new BoxRegion();
		_mapMarkers = new BoxRegion();
		_learning = false;
		_mapping = new HashMap<Integer, MapControl>();
		_invMapping = new HashMap<String, Integer>();
		_highlighter = null;
		_mousedown = false;
		_chasing = true;
		_selectedBoxes = new LinkedList<String>();
		_highlightMap = new HashMap<String,MaxBox>();
		_lastDisplayed = null;
		_blockRepeatMap = false;
		_mouseLocation = new int[]{0,0};
		_devices = new HashSet<Character>();
		_delayDisplay = new MaxClock(this, "setdisplay");
		
	}

	public void loadbang() {
		outlet(1, "dump");
	}
	
	//get position of mouse and highlight box under mouse (during learning)
	public void mousepos(int x, int y) {
		//don't select anything if mouse button is already down
		if ( _mousedown )
			return;
		int[] winloc = _myWindow.getLocation();
		_mouseLocation[0] = x - winloc[0];
		_mouseLocation[1] = y - winloc[1];
		MaxBox newselection = _boxes.GetSelection(_mouseLocation[0], _mouseLocation[1]);
		if (newselection != null){
			if (_highlightedBox == newselection) {
				//do nothing already highlighted
			}
			else if (_highlightMap.containsKey(newselection.getName())) {
				//do nothing box is part of multiple selection
			}
			else { //set new highlighter and associated box
				_highlightedBox = newselection;
				int[] rec = _highlightedBox.getRect();
				rec[0] -= 2;
				rec[1] -= 2;
				rec[2] += 2;
				rec[3] += 2;
				if ( _highlighter != null) { //remove previous highlight
					_highlighter.remove();
				}
				if  (_learning) { //set up the highlighter
					_highlighter = _myParent.newDefault(rec[0], rec[1], "panel", null);
					_highlighter.setIgnoreClick(true);
					_highlighter.setRect(rec[0], rec[1], rec[2], rec[3] );
					_highlighter.setName("mmhighlighter");
					_highlighter.send("bgcolor", new Atom[]{ 
							Atom.newAtom(255), Atom.newAtom(255), Atom.newAtom(0), Atom.newAtom(0)});
					_highlighter.send("frgb", new Atom[]{ 
							Atom.newAtom(255), Atom.newAtom(255), Atom.newAtom(0)});
					_highlighter.send("border", new Atom[]{Atom.newAtom(3)});
					outlet(2, new Atom[]{Atom.newAtom("script"), Atom.newAtom("bringtofront"), 
							Atom.newAtom(_highlighter.getName())});
				}
			}
		}
		else { //mouse not over box, remove highlighter if it exists
			if (_highlightedBox != null) {
				if ( _highlighter != null) {
					_highlighter.remove();
				}
				_highlightedBox = null;
			}
			MaxBox mapmark = _mapMarkers.GetSelection(_mouseLocation[0], _mouseLocation[1]);
			if (_lastMarker != null) {
				if (mapmark != _lastMarker) {
					if (! _lastMarker.equals(_selectedMarker)) {
						_lastMarker.send("bgcolor", new Atom[]{ 
								Atom.newAtom(.6), Atom.newAtom(1), Atom.newAtom(.6), Atom.newAtom(1)});
					}
					if(mapmark != null) {
						_lastMarker = mapmark;
						_lastMarker.send("bgcolor", new Atom[]{ 
								Atom.newAtom(.3), Atom.newAtom(1), Atom.newAtom(.3), Atom.newAtom(1)});
					}
					else {
						_lastMarker = null;
					}
				}
			}
			else {
				if(mapmark != null) {
					_lastMarker = mapmark;
					_lastMarker.send("bgcolor", new Atom[]{ 
							Atom.newAtom(.3), Atom.newAtom(1), Atom.newAtom(.3), Atom.newAtom(1)});
				}
			}
		}
	
	}
	
	//get state of mouse button
	public void mousebut(int b) {
		if (b == 1)
			_mousedown = true;
		else
			_mousedown = false;
		//if learning and mouse click
		if (_learning && (_mousedown)) {
			//if there is a highlighted box
			if ((_highlightedBox != null) && (_highlighter != null))  {
				// if box is not selected, select
				if (! _highlightMap.containsKey(_highlightedBox.getName())){
					_highlighter.send("frgb", new Atom[]{ 
							Atom.newAtom(0), Atom.newAtom(255), Atom.newAtom(0)});
					_highlighter.send("bgcolor", new Atom[]{ 
							Atom.newAtom(255), Atom.newAtom(255), Atom.newAtom(0), Atom.newAtom(0)});
					_highlightMap.put(_highlightedBox.getName(),_highlighter);
					_selectedBoxes.add(_highlightedBox.getName());
					_highlighter = null;
					_highlightedBox = null;
					
				}
				else { //if box is was already selected, deselect all
					removeselection();
				}
			}
			else { //remove all selections on click on nothing
					removeselection();
					MaxBox marker = _mapMarkers.GetSelection(_mouseLocation[0], _mouseLocation[1]);
					if (marker != null) {
						if (_invMapping.containsKey(marker.getName().substring(6))) {
							_lastHash = _invMapping.get(marker.getName().substring(6));
							if (_selectedMarker != null) {
								if ( ! _selectedMarker.equals(marker)) {
									_selectedMarker.send("bgcolor", new Atom[]{
											Atom.newAtom(.6), Atom.newAtom(1), Atom.newAtom(.6), Atom.newAtom(1)});
								}
							}
							_selectedMarker = marker;
							queuedisplay(_lastHash);
						}
					}
			}
		}
	}
	
	//remove all selected boxes and associated highlighters
	private void removeselection() {
		Collection<MaxBox> highlights = _highlightMap.values();
		Iterator<MaxBox> nexthighlight = highlights.iterator();
		while ( nexthighlight.hasNext()) {
			nexthighlight.next().remove();
			nexthighlight.remove();
		}
		_selectedBoxes.clear();
		_blockRepeatMap = false;
	}
	
	
	//key pressed (ascii value)
	public void key(char k) {
		int thishash = (int)k;
		// if learning and there is at least one selected box
		if (_learning && (_selectedBoxes.size() > 0)) {
			//prevent a single key from mapping to multiple boxes in a selection
			if ((_lastHash != thishash) || ! _blockRepeatMap || (! _mapping.containsKey(thishash))) {
				String name = _selectedBoxes.removeFirst();
				_highlightMap.remove(name).remove(); //remove highlight on learn
				
				//only map one-to-one (remove previous mapping)
				if (_invMapping.containsKey(name)) {
					MaxBox marker = _myParent.getNamedBox("marker"+name);
					if (marker != null) {
						//_markers.remove(marker.getName());
						_mapMarkers.RemoveBox(marker);
						marker.remove();
					}
					int oldhash = _invMapping.remove(name);
					_mapping.remove(oldhash);
					outlet(1, new Atom[]{Atom.newAtom("remove"), Atom.newAtom(oldhash)});
					
				}
				if (_mapping.containsKey(thishash)) {
					MaxBox marker = _myParent.getNamedBox("marker"+_mapping.get(thishash).name());
					if (marker != null) {
						//_markers.remove(marker.getName());
						_mapMarkers.RemoveBox(marker);
						marker.remove();
					}
					_invMapping.remove(_mapping.remove(thishash).name());
					//outlet(1, new Atom[]{Atom.newAtom("remove"), Atom.newAtom(thishash)});
				}
				//make sure box still exists in patcher
				if (_myParent.getNamedBox(name) == null) {
					post("error mapping box: " + name + " DNE\n");
					return;
				}
				//use settings of last displayed map setting if it is of the same type
				if (_lastDisplayed != null) {
					if (_lastDisplayed.type() == MapControl.KEY) {
						_mapping.put(thishash, new KeyMapControl(_myParent.getNamedBox(name), name, MapControl.KEY, 
								_lastDisplayed.min(), _lastDisplayed.max(), _lastDisplayed.precision(), _lastDisplayed.curve(),
								_lastDisplayed.sendUpOff(), _lastDisplayed.sendVelocity(), _lastDisplayed.banger()));
					}
					else { //use default settings
						_mapping.put(thishash, new KeyMapControl(_myParent.getNamedBox(name), name, 
								MapControl.KEY, 0, 127));
					}
				}
				else { //use default settings
					_mapping.put(thishash, new KeyMapControl(_myParent.getNamedBox(name), name, 
							MapControl.KEY, 0, 127));
				}
				_invMapping.put(name, thishash);
				outputsetting(thishash);
				addmarker(_myParent.getNamedBox(name));
				if (_selectedBoxes.size() > 0)
					_blockRepeatMap = true;
				else
					_blockRepeatMap = false;
			}
		}
		//if not learning send control value to the mapped box
		if (_mapping.containsKey(thishash)) {
			if (_chasing) { //display settings for this mapping if chasing
				//if this box does not exist remove it from the mapping
				if (_myParent.getNamedBox(_mapping.get(thishash).name()) == null) {
					_invMapping.remove(_mapping.remove(thishash).name());
					queuedisplay(_lastHash);
					return;
				}
				queuedisplay(thishash);
			}
			_mapping.get(thishash).send(127);
		}
		else if (_chasing) {
			queuedisplay(thishash);
		}
		_lastHash = thishash;
	}
	
	//key released
	public void keyup(char k) {
		_lastHash = (int)k;
		if (_mapping.containsKey(_lastHash)) {
			_mapping.get(_lastHash).send(0);
		}
	}
	
	private int hashmidi(char dev, int type, int channel, int dataone) {
		int thishash = ((int)dev*16+type)*16+channel;
		switch (type) {
		case 8: //note off
			type = 9; //treat as note on with 0 velocity
			thishash = ((int)dev*16+type)*16+channel;
		case 9: //note on
		case 10: //key pressure
		case 11: //controller change
			thishash = thishash*128+dataone;
			break;
		case 12: //program change (dataone only)
		case 13: //channel pressure (dataone only)
			thishash = thishash*128;
			break;
		case 14: //pitch bend
			thishash = thishash*128;
			break;
		}
		return thishash;
	}
	
	private int[] unhashmidi(int hash) {
		int[] unhash = new int[4];
		unhash[3] = hash & 127; //dataone
		unhash[2] = (hash>>7) & 15; //channel
		unhash[1] = (hash>>11) & 15; //type
		unhash[0] = (hash>>15) & 255; //dev
		return unhash;
	}
	
	//receive midi input
	public void midi(char dev, int type, int channel, int dataone, int datatwo) {
		int thishash = ((int)dev*16+type)*16+channel;
		//_lastHash = ((int)dev*16+type)*16+channel;
		int contype = MapControl.CONTROLLER;
		int val = datatwo;
		switch (type) {
		case 8: //note off
			type = 9; //treat as note on with 0 velocity
			datatwo = 0;
			val = datatwo;
			thishash = ((int)dev*16+type)*16+channel;
		case 9: //note on
			contype = MapControl.NOTE;
		case 10: //key pressure
			thishash = thishash*128+dataone;
			break;
		case 11: //controller change
			contype = MapControl.CC;
			thishash = thishash*128+dataone;
			break;
		case 12: //program change (dataone only)
		case 13: //channel pressure (dataone only)
			thishash = thishash*128;
			val = dataone;
			break;
		case 14: //pitch bend
			thishash = thishash*128;
			val = datatwo;
			break;
		}
		if (_learning && (_selectedBoxes.size() > 0)) {
			//don't repeatedly map same midi controller
			if ((_lastHash != thishash) || ! _blockRepeatMap || (! _mapping.containsKey(thishash))) {
				String name = _selectedBoxes.removeFirst();
				_highlightMap.remove(name).remove(); //remove highlight on learn
				if (_invMapping.containsKey(name)) {
					MaxBox marker = _myParent.getNamedBox("marker"+name);
					if (marker != null) {
						//_markers.remove(marker.getName());
						_mapMarkers.RemoveBox(marker);
						marker.remove();
					}
					int oldhash = _invMapping.remove(name);
					_mapping.remove(oldhash);
					outlet(1, new Atom[]{Atom.newAtom("remove"), Atom.newAtom(oldhash)});
				}
				if (_mapping.containsKey(thishash)) {
					MaxBox marker = _myParent.getNamedBox("marker"+_mapping.get(thishash).name());
					if (marker != null) {
						//_markers.remove(marker.getName());
						_mapMarkers.RemoveBox(marker);
						marker.remove();
					}
					_invMapping.remove(_mapping.remove(thishash).name());
				}
				if (_myParent.getNamedBox(name) == null) {
					post("error mapping box: " + name + " DNE\n");
					return;
				}
				if (_lastDisplayed != null) {
					if (_lastDisplayed.type() == contype) {
						if (contype == MapControl.CC) {
							_mapping.put(thishash, new CcMapControl(_myParent.getNamedBox(name), name, contype, 
									_lastDisplayed.min(), _lastDisplayed.max(), _lastDisplayed.precision(), 
									_lastDisplayed.curve(), ((CcMapControl)_lastDisplayed).encoder(), 
									_lastDisplayed.sendUpOff(),_lastDisplayed.sendVelocity(), _lastDisplayed.banger()));
						}
						else {
							_mapping.put(thishash, new MapControl(_myParent.getNamedBox(name), name, contype, 
								_lastDisplayed.min(), _lastDisplayed.max(), _lastDisplayed.precision(), _lastDisplayed.curve(),
								_lastDisplayed.sendUpOff(), _lastDisplayed.sendVelocity(), _lastDisplayed.banger()));
						}
					}
					else if (contype == MapControl.CC) {
						_mapping.put(thishash, new CcMapControl(_myParent.getNamedBox(name), name, contype, 0, 127));
					}
					else {
						_mapping.put(thishash, new MapControl(_myParent.getNamedBox(name), name, contype, 0, 127));
					}
				}
				else if (contype == MapControl.CC)
					_mapping.put(thishash, new CcMapControl(_myParent.getNamedBox(name), name, contype, 0, 127));
				else
					_mapping.put(thishash, new MapControl(_myParent.getNamedBox(name), name, contype, 0, 127));
				_invMapping.put(name, thishash);
				outputsetting(thishash);
				addmarker(_myParent.getNamedBox(name));
				
				if (_selectedBoxes.size() > 0)
					_blockRepeatMap = true;
				else
					_blockRepeatMap = false;
			}
		}
		if (_mapping.containsKey(thishash)) {
			if (_chasing) {
				try {
					if (_myParent.getNamedBox(_mapping.get(thishash).name()) == null) {
						_invMapping.remove(_mapping.remove(thishash).name());
						queuedisplay(thishash);
						return;
					}
				} catch (Exception e) {
					post("chase exception");
					return;
				}
				queuedisplay(thishash);
			}
			_mapping.get(thishash).send(val);
		}
		else if (_chasing) {
			queuedisplay(thishash);
		}
		_lastHash = thishash;
			
	}
	
	//enable midi devices (from textedit box)
	public void enabledevices(String string) {
		//delete devices if bang received (empty textedit box)
		if (string.equals("bang")) {
			Iterator<Character> olddevs = _devices.iterator();
			while (olddevs.hasNext()) {
				Character dev = olddevs.next();
				outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
						Atom.newAtom("delete"), Atom.newAtom("midiin"+dev)});
				outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
						Atom.newAtom("delete"), Atom.newAtom("midigroup"+dev)});
				outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
						Atom.newAtom("delete"), Atom.newAtom("prepend"+dev)});
				olddevs.remove();
			}
		}
		else {
			Set<Character> newdevices = new HashSet<Character>();
			char[] chars = string.toCharArray();
			//_devices.clear();
			//create new devices
			for (int i = 0; i < chars.length; i++) {
				if (Character.isLetter(chars[i])) {
					if (! _devices.contains(chars[i])) {
						if (! newdevices.contains(chars[i])) {
							newdevices.add(Character.toLowerCase(chars[i]));
							outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"), 
									Atom.newAtom("newdefault"), Atom.newAtom("midiin"+chars[i]), 
									Atom.newAtom(10+i*100), Atom.newAtom(100), 
									Atom.newAtom("midiin"), Atom.newAtom(chars[i])});
							outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"), 
									Atom.newAtom("newdefault"), Atom.newAtom("midigroup"+chars[i]),
									Atom.newAtom(10+i*100), Atom.newAtom(150), 
									Atom.newAtom("midiGroup3.maxpat")});
							outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"), 
									Atom.newAtom("newdefault"), Atom.newAtom("prepend"+chars[i]),
									Atom.newAtom(10+i*100), Atom.newAtom(190), 
									Atom.newAtom("prepend"), Atom.newAtom("midi"), Atom.newAtom(chars[i])});
							outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
									Atom.newAtom("connect"), Atom.newAtom("midiin"+chars[i]), Atom.newAtom(0),
									Atom.newAtom("midigroup"+chars[i]), Atom.newAtom(0)});
							outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
									Atom.newAtom("connect"), Atom.newAtom("midigroup"+chars[i]), Atom.newAtom(0),
									Atom.newAtom("prepend"+chars[i]), Atom.newAtom(0)});
							outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
									Atom.newAtom("connect"), Atom.newAtom("prepend"+chars[i]), Atom.newAtom(0),
									Atom.newAtom("patchoutlet"), Atom.newAtom(0)});
						}
					}
					else {
						_devices.remove(chars[i]);
						newdevices.add(chars[i]);
					}
				}
			}
			//delete old devices
			Iterator<Character> olddevs = _devices.iterator();
			while (olddevs.hasNext()) {
				Character dev = olddevs.next();
				outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
						Atom.newAtom("delete"), Atom.newAtom("midiin"+dev)});
				outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
						Atom.newAtom("delete"), Atom.newAtom("midigroup"+dev)});
				outlet(0, new Atom[]{Atom.newAtom("midipatcher"), Atom.newAtom("script"),
						Atom.newAtom("delete"), Atom.newAtom("prepend"+dev)});
			}
			_devices = newdevices;
		}
		
		Iterator<Character> iter = _devices.iterator();
		String devs = "";
		while (iter.hasNext()) {
			char nextchar = iter.next();
			devs = devs + nextchar;
		}
		//save midi device settings in coll
		outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom("mididevices"), Atom.newAtom(devs)});
	}

	//turn learning on/off
	public void learn(int l) {
		if (l == 0) {
			_learning = false;
			outlet(0, "untrackmouse");
			if (_highlighter != null)
				_highlighter.remove();
			_highlightedBox = null;
			_mousedown = false;
			_mapMarkers.RemoveAllBoxes();
			//_markers.clear();
			//removemarkers();
			_selectedMarker = null;
			removeselection();
		}
		else {
			checkmap();
			loadboxes();
			displaymarkers();
			outlet(0, "trackmouse");
			_learning = true;
			_highlightedBox = null;
			if (_chasing)
				queuedisplay(_lastHash);
		}
	}
	
	//display indicator for mapped box
	private void addmarker(MaxBox mapbox) {
		MaxBox marker;
		int[] rec = mapbox.getRect();
		marker = _myParent.newDefault(rec[0], rec[1], "panel", null);
		marker.setRect(rec[0]-4, rec[1]-8, rec[0]+14, rec[1]);
		marker.send("bgcolor", new Atom[]{ 
				Atom.newAtom(.6), Atom.newAtom(1), Atom.newAtom(.6), Atom.newAtom(1)});
		marker.setName("marker"+mapbox.getName());
		marker.send("rounded", new Atom[]{Atom.newAtom(0)});
		outlet(2, new Atom[]{Atom.newAtom("script"), Atom.newAtom("bringtofront"), 
				Atom.newAtom(marker.getName())});
		_mapMarkers.AddBox(marker);
		//_markers.add(marker.getName());
	}
	
	//display indicator for each mapped box
	private void displaymarkers() {
		Iterator<MapControl> settings;
		Collection<MapControl> allsettings = _mapping.values();
		settings = allsettings.iterator();
		while(settings.hasNext()) {
			MaxBox mapbox = settings.next().theBox();
			addmarker(mapbox);
		}
	}
	
	//check for and try to correct errors in mapping (boxes deleted, name changes)
	private void checkmap() {
		Iterator<MapControl> settings;
		Collection<MapControl> allsettings = _mapping.values();
		settings = allsettings.iterator();
		while(settings.hasNext()) {
			MapControl next = settings.next();
			MaxBox namedBox = _myParent.getNamedBox(next.name());
				if (namedBox != null) {
					if (namedBox.equals(next.theBox())){
					}
					else {
						try {
							if (_myParent.getNamedBox(next.theBox().getName()) != null) { 
								//set to box's new name
								int hash = _invMapping.remove(next.name());
								next.name(next.theBox().getName());
								_invMapping.put(next.name(), hash);
								if (namedBox.equals(next.theBox())){
									outputsetting(hash);
								}
								else {
									next.theBox(namedBox);
									outputsetting(hash);
								}
							}
							else { //set to box with same name
								next.theBox(namedBox);
								outputsetting(_invMapping.get(next.name()));
							}
						} catch (Exception e) {
							next.theBox(namedBox);
							outputsetting(_invMapping.get(next.name()));
						}
					}
				}
				else {
					namedBox = _myParent.getNamedBox(next.theBox().getName());
					if (namedBox != null) { //set to box's new name
						int hash = _invMapping.remove(next.name());
						next.name(namedBox.getName());
						_invMapping.put(next.name(), hash);
						outputsetting(hash);
					}
					else {
						//remove
						int hash = _invMapping.remove(next.name());
						settings.remove();
						outlet(1, new Atom[]{Atom.newAtom("remove"), Atom.newAtom(hash)});
					}
				}
			}
		
	}

	//load all boxes so they can be selected for mapping
	private void loadboxes()
	{
		_boxes.RemoveAll();
		//_includedBoxes.clear();
		MaxBox[] allboxes = _myParent.getAllBoxes();
		for( int i = 0; i < allboxes.length; i++ )
		{
			//disable mapping for special boxes
			if ((!allboxes[i].equals(_myBox)) && (!allboxes[i].equals(_guiBox)) 
					&& ! allboxes[i].getName().equals("mmapPrepend")
					&& ! allboxes[i].getName().equals("mmapColl")
					&& ! allboxes[i].getName().equals("mmapDump")
					&& ! allboxes[i].getName().equals("mmapThisPatcher")
					&& ! allboxes[i].getMaxClass().equals("panel") 
					&& ! allboxes[i].getMaxClass().equals("comment")) {
				try {
					if (allboxes[i].getName().length() == 0) {
						allboxes[i].setName("mm"+allboxes[i].getMaxClass() + i);
					}
				} catch (Exception e) {
					allboxes[i].setName("mm"+allboxes[i].getMaxClass() + i);
					post("exception found!\n");
				}
				_boxes.AddBox(allboxes[i]);
				//_includedBoxes.add(allboxes[i].getName());
			}
		}
	}
	
	//output settings for key/midi to coll
	private void outputsetting(int hash) {
		MapControl setting = _mapping.get(hash);
		if (setting != null) {
			if (setting.type() == MapControl.CC) {
				try {
				outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom(hash), 
						Atom.newAtom(hash), Atom.newAtom(setting.name()), 
						Atom.newAtom(setting.type()), Atom.newAtom(setting.min()), Atom.newAtom(setting.max()), 
						Atom.newAtom(setting.precision()), Atom.newAtom(setting.curve()), 
						Atom.newAtom(((CcMapControl)setting).encoder()), Atom.newAtom(setting.sendUpOff()), 
						Atom.newAtom(setting.sendVelocity()), Atom.newAtom(setting.banger())});
				} catch (Exception e) {
					outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom(hash), 
							Atom.newAtom(hash), Atom.newAtom(setting.name()), 
							Atom.newAtom(setting.type()),Atom.newAtom(setting.min()), Atom.newAtom(setting.max()), 
							Atom.newAtom(setting.precision()), Atom.newAtom(setting.curve()), 
							Atom.newAtom(setting.sendUpOff()), 
							Atom.newAtom(setting.sendVelocity()), Atom.newAtom(setting.banger())});
				}
			}
			else
				outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom(hash), 
					Atom.newAtom(hash), Atom.newAtom(setting.name()), 
					Atom.newAtom(setting.type()),Atom.newAtom(setting.min()), Atom.newAtom(setting.max()), 
					Atom.newAtom(setting.precision()), Atom.newAtom(setting.curve()), Atom.newAtom(setting.sendUpOff()), 
					Atom.newAtom(setting.sendVelocity()), Atom.newAtom(setting.banger())});
		}
	}
	
	public void dumpsetting() {
		//if last touched is mapped
		if (_mapping.containsKey(_lastHash)) {
			MapControl setting = _mapping.get(_lastHash);
			if (setting != null) {
				int[] loc = _myBox.getRect();
				//loc[1] -= 20;
				loc[0] += 100;
				MaxBox b1 = _myParent.newDefault(loc[0], loc[1], "prepend", new Atom[]{Atom.newAtom("edit"),
						Atom.newAtom(_lastHash), Atom.newAtom("banger")});
				_myParent.connect(b1, 0, _myBox, 0);
				b1 = _myParent.newDefault(loc[0], loc[1]-20, "prepend", new Atom[]{Atom.newAtom("edit"),
						Atom.newAtom(_lastHash), Atom.newAtom("min")});
				_myParent.connect(b1, 0, _myBox, 0);
				b1 = _myParent.newDefault(loc[0], loc[1]-40, "prepend", new Atom[]{Atom.newAtom("edit"),
						Atom.newAtom(_lastHash), Atom.newAtom("max")});
				_myParent.connect(b1, 0, _myBox, 0);
				if (setting.type() == MapControl.KEY) {
					b1 = _myParent.newDefault(loc[0], loc[1]-60, "prepend", new Atom[]{Atom.newAtom("edit"),
							Atom.newAtom(_lastHash), Atom.newAtom("keyup")});
					_myParent.connect(b1, 0, _myBox, 0);
				}
				else {
					b1 = _myParent.newDefault(loc[0], loc[1]-60, "prepend", new Atom[]{Atom.newAtom("edit"),
							Atom.newAtom(_lastHash), Atom.newAtom("precision")});
					_myParent.connect(b1, 0, _myBox, 0);
					b1 = _myParent.newDefault(loc[0], loc[1]-80, "prepend", new Atom[]{Atom.newAtom("edit"),
							Atom.newAtom(_lastHash), Atom.newAtom("curve")});
					_myParent.connect(b1, 0, _myBox, 0);
					if (setting.type() == MapControl.NOTE) {
						b1 = _myParent.newDefault(loc[0], loc[1]-100, "prepend", new Atom[]{Atom.newAtom("edit"),
								Atom.newAtom(_lastHash), Atom.newAtom("noteoff")});
						_myParent.connect(b1, 0, _myBox, 0);
						b1 = _myParent.newDefault(loc[0], loc[1]-120, "prepend", new Atom[]{Atom.newAtom("edit"),
								Atom.newAtom(_lastHash), Atom.newAtom("velocity")});
						_myParent.connect(b1, 0, _myBox, 0);
					}
					else { 
						b1 = _myParent.newDefault(loc[0], loc[1]-100, "prepend", new Atom[]{Atom.newAtom("edit"),
								Atom.newAtom(_lastHash), Atom.newAtom("sendmin")});
						_myParent.connect(b1, 0, _myBox, 0);
						/*if (setting.type() == MapControl.CC) {
							b1 = _myParent.newDefault(loc[0], loc[1]-120, "prepend", new Atom[]{Atom.newAtom("edit"),
								Atom.newAtom(_lastHash), Atom.newAtom("encoder")});
							_myParent.connect(b1, 0, _myBox, 0);
						}*/
					}
				}
			}
		}
	}
	
	//unmap current selection
	public void unmap() {
		if (_mapping.containsKey(_lastHash)) {
			MapControl removed = _mapping.remove(_lastHash);
			_invMapping.remove(removed.name());
			MaxBox marker = _myParent.getNamedBox("marker"+removed.name());
			if (marker != null) {
				_mapMarkers.RemoveBox(marker);
				//_markers.remove(marker.getName());
				marker.remove();
			}
			//remove setting from coll
			outlet(1, new Atom[]{Atom.newAtom("remove"), Atom.newAtom(_lastHash)});
			queuedisplay(_lastHash);
		}
	}
	
	//toggle chasing (display of last key/midi touched)
	public void chase(int c) {
		if (c == 0) {
			_chasing = false;
			if (! _learning || (_selectedMarker == null)) {
				outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
						Atom.newAtom("sendbox"), Atom.newAtom("keyGUI"), 
						Atom.newAtom("presentation_position"), Atom.newAtom(400), Atom.newAtom(0)});
				outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
						Atom.newAtom("sendbox"), Atom.newAtom("noteGUI"), 
						Atom.newAtom("presentation_position"), Atom.newAtom(800), Atom.newAtom(0)});
				outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
						Atom.newAtom("sendbox"), Atom.newAtom("controllerGUI"), 
						Atom.newAtom("presentation_position"), Atom.newAtom(600), Atom.newAtom(0)});
				outlet(0, new Atom[]{Atom.newAtom("mapped"), Atom.newAtom(0)});
				_lastDisplayed = null;
			}
		}
		else
			_chasing = true;
			
	}
	
	//set values of settings in the gui for key/midi mapping
	private void updatedisplay(MapControl control, int hash) {
		switch (control.type()) {
		case MapControl.CONTROLLER:
			outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
					Atom.newAtom("encoder"), Atom.newAtom(-1)});
		case MapControl.CC:
			try {
				outlet(0, new Atom[]{Atom.newAtom("controllerGUI"),
					Atom.newAtom("encoder"), Atom.newAtom(((CcMapControl)control).encoder())});
				} catch (Exception e) {}
			outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
					Atom.newAtom("min"), Atom.newAtom(control.min())});
			outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
					Atom.newAtom("max"), Atom.newAtom(control.max())});
			outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
					Atom.newAtom("precision"), Atom.newAtom(control.precision())});
			outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
					Atom.newAtom("curve"), Atom.newAtom(control.curve())});
			switch (hash>>11 & 0xF) {
			case 0xA: outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), Atom.newAtom("display"), 
					Atom.newAtom((char)(hash>>15) + " Poly press. \nKey: " + (hash & 0x7F) + " ch: " + (((hash>>7) & 0xF) + 1))});
				break;
			case 0xB: outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), Atom.newAtom("display"), 
					Atom.newAtom((char)(hash>>15) + " Cctrl: " + (hash & 0x7F) + " \nch: " + (((hash>>7) & 0xF) + 1))});
				break;
			case 0xC: outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), Atom.newAtom("display"), 
					Atom.newAtom((char)(hash>>15) + " Prg Change \nch: " + (((hash>>7) & 0xF) + 1))});
				break;
			case 0xD: 
					outlet(0, new Atom[]{Atom.newAtom("controllerGUI"),Atom.newAtom("display"), 
					Atom.newAtom((char)(hash>>15) + " After touch \nch: " + ((hash>>7) & 0xF) + 1)});
				break;
			case 0xE: outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), Atom.newAtom("display"), 
					Atom.newAtom((char)(hash>>15) + " Pitch bend \nch: " + (((hash>>7) & 0xF) + 1))});
				break;
			default: outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
							Atom.newAtom("display"), Atom.newAtom(hash & 127)});
			}
			
			outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
					Atom.newAtom("banger"), Atom.newAtom(control.banger())});
			outlet(0, new Atom[]{Atom.newAtom("controllerGUI"), 
					Atom.newAtom("noteoff"), Atom.newAtom(control.sendUpOff())});
			break;
		case MapControl.NOTE:
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), 
					Atom.newAtom("noteoff"), Atom.newAtom(control.sendUpOff())});
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), 
					Atom.newAtom("velocity"), Atom.newAtom(control.sendVelocity())});
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), 
					Atom.newAtom("min"), Atom.newAtom(control.min())});
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), 
					Atom.newAtom("max"), Atom.newAtom(control.max())});
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), 
					Atom.newAtom("precision"), Atom.newAtom(control.precision())});
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), 
					Atom.newAtom("curve"), Atom.newAtom(control.curve())});
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), Atom.newAtom("display"), 
					Atom.newAtom((char)(hash>>15) + " note: " + (hash & 0x7F) + "  ch: " + (((hash>>7) & 0xF) + 1))});
			outlet(0, new Atom[]{Atom.newAtom("noteGUI"), 
					Atom.newAtom("banger"), Atom.newAtom(control.banger())});
			break;
		case MapControl.KEY:
			outlet(0, new Atom[]{Atom.newAtom("keyGUI"), Atom.newAtom("keyup"), 
					Atom.newAtom(control.sendUpOff())});
			outlet(0, new Atom[]{Atom.newAtom("keyGUI"), 
					Atom.newAtom("min"), Atom.newAtom(control.min())});
			outlet(0, new Atom[]{Atom.newAtom("keyGUI"), 
					Atom.newAtom("max"), Atom.newAtom(control.max())});
			String special = "";
			switch (hash) {
				case 1: special = "Home";
				break;
				case 4: special = "End";
				break;
				case 8: special = "BackSpc";
				break;
				case 11: special = "PgUp";
				break;
				case 12: special = "PgDn";
				break;
				case 27: special = "Esc";
				break;
				case 28: special = "Left";
				break;
				case 29: special = "Right";
				break;
				case 30: special = "Up";
				break;
				case 31: special = "Down";
				break;
				case 32: special = "Space";
				break;
				case 127: special = "Del";
				break;
			}
			if (special.equals("")) {
				outlet(0, new Atom[]{Atom.newAtom("keyGUI"), 
						Atom.newAtom("display"), Atom.newAtom((char)hash)});
			}
			else {
				outlet(0, new Atom[]{Atom.newAtom("keyGUI"), 
						Atom.newAtom("display"), Atom.newAtom(special)});
			}
			outlet(0, new Atom[]{Atom.newAtom("keyGUI"), 
					Atom.newAtom("banger"), Atom.newAtom(control.banger())});
			break;
		}
	}
	
	//display settings for hash in 20ms (prevents cpu overload when moving multiple controllers)
	public void queuedisplay(int hash) {		
		_queueHash = hash;
		_delayDisplay.delay(20);
	}
	
	//display (allow for editing) of settings of key/controller with specified hash
	public void setdisplay() {
		
		MapControl control = _mapping.get(_queueHash);
		
		//nothing mapped to this control yet, allow for editing of setting before mapping
		if (control == null) {
			int[] unhash = unhashmidi(_queueHash);
			int ctype = MapControl.CONTROLLER;
			if (unhash[1] == 9)
				ctype = MapControl.NOTE;
			else if (unhash[1] == 11)
				ctype = MapControl.CC;
			if(_queueHash < 1024) {
				ctype = MapControl.KEY;
				if ((_lastDisplayed != null) && (_lastDisplayed.type() == ctype)) {
					control = new KeyMapControl(null, null, ctype, _lastDisplayed.min(), _lastDisplayed.max(),
							_lastDisplayed.precision(), _lastDisplayed.curve(), _lastDisplayed.sendUpOff(), 
							_lastDisplayed.sendVelocity(), _lastDisplayed.banger());
				}
				else {
					control = new KeyMapControl(null, null, ctype, 0, 127);
				}
			}
			else if ((_lastDisplayed != null) && (_lastDisplayed.type() == ctype)) {
				if (ctype == MapControl.CC) {
					control = new CcMapControl(null, null, ctype, _lastDisplayed.min(), _lastDisplayed.max(),
							_lastDisplayed.precision(), _lastDisplayed.curve(), ((CcMapControl)_lastDisplayed).encoder(),
							_lastDisplayed.sendUpOff(), _lastDisplayed.sendVelocity(), _lastDisplayed.banger());
				}
				else {
					control = new MapControl(null, null, ctype, _lastDisplayed.min(), _lastDisplayed.max(),
						_lastDisplayed.precision(), _lastDisplayed.curve(), _lastDisplayed.sendUpOff(), 
						_lastDisplayed.sendVelocity(), _lastDisplayed.banger());
				}
			}
			else {
				if (ctype == MapControl.CC)
					control = new CcMapControl(null, null, ctype, 0, 127);
				else
					control = new MapControl(null, null, ctype, 0, 127);
			}
			if (_selectedMarker != null) {
				if (_myParent.getNamedBox(_selectedMarker.getName()) != null ) {
					_selectedMarker.send("bgcolor", new Atom[]{ 
							Atom.newAtom(.6), Atom.newAtom(1), Atom.newAtom(.6), Atom.newAtom(1)});
				}
				_selectedMarker = null;
			}
			outlet(0, new Atom[]{Atom.newAtom("mapped"), Atom.newAtom(0)});
		}
		else
			outlet(0, new Atom[]{Atom.newAtom("mapped"), Atom.newAtom(1)});
		
		//else 
		if (control == _lastDisplayed) {
			//do nothing
			if (_learning && _selectedMarker == null) {
				MaxBox marker = _myParent.getNamedBox("marker"+control.name());
				if (marker != null) {
					_selectedMarker = marker;
					_selectedMarker.send("bgcolor", new Atom[]{ 
							Atom.newAtom(0), Atom.newAtom(1), Atom.newAtom(0), Atom.newAtom(1)});
				}
			}
		}
		else {
			if (_learning) {
				//change color of markers to indicate which maxbox's mapping settings are being displayed
				if (_selectedMarker != null) {
					try {
						int mhash = _invMapping.get(_selectedMarker.getName().substring(6));
						if (_queueHash != mhash) {
							_selectedMarker.send("bgcolor", new Atom[]{ 
									Atom.newAtom(.6), Atom.newAtom(1), Atom.newAtom(.6), Atom.newAtom(1)});
							_selectedMarker = null;
							MaxBox marker = _myParent.getNamedBox("marker"+control.name());
							if (marker != null) {
								_selectedMarker = marker;
							}
						}
					} catch (Exception e) {
						_selectedMarker = null;
						//post("selectedmarker = null catch");
					}
				}
				else {
					MaxBox marker = _myParent.getNamedBox("marker"+control.name());
					if (marker != null) {
						_selectedMarker = marker;
					}
				}
				if (_selectedMarker != null) {
					_selectedMarker.send("bgcolor", new Atom[]{ 
							Atom.newAtom(0), Atom.newAtom(1), Atom.newAtom(0), Atom.newAtom(1)});
				}
			}
			if (_lastDisplayed == null) {
				_lastDisplayed = control;
			}
			else if (control.type() == _lastDisplayed.type()){
				_lastDisplayed = control;
			}
			else {	//move last GUI out of focus 
				switch (_lastDisplayed.type()) {
				case MapControl.CC:
					if (control.type() == MapControl.CONTROLLER)
						break;
				case MapControl.CONTROLLER:
					if (control.type() == MapControl.CC)
						break;
					outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
							Atom.newAtom("sendbox"), Atom.newAtom("controllerGUI"), 
							Atom.newAtom("presentation_position"), Atom.newAtom(600), Atom.newAtom(0)});
					break;
				case MapControl.KEY:
					outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
							Atom.newAtom("sendbox"), Atom.newAtom("keyGUI"), 
							Atom.newAtom("presentation_position"), Atom.newAtom(400), Atom.newAtom(0)});
					break;
				case MapControl.NOTE:
					outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
							Atom.newAtom("sendbox"), Atom.newAtom("noteGUI"), 
							Atom.newAtom("presentation_position"), Atom.newAtom(800), Atom.newAtom(0)});
					break;
				}
				_lastDisplayed = control;
			}
			//move correct GUI into focus
			switch (_lastDisplayed.type()) {
			case MapControl.CC:
			case MapControl.CONTROLLER:
				outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
						Atom.newAtom("sendbox"), Atom.newAtom("controllerGUI"), 
						Atom.newAtom("presentation_position"), Atom.newAtom(69), Atom.newAtom(0)});
				break;
			case MapControl.KEY:
				outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
						Atom.newAtom("sendbox"), Atom.newAtom("keyGUI"), 
						Atom.newAtom("presentation_position"), Atom.newAtom(69), Atom.newAtom(0)});
				break;
			case MapControl.NOTE:
				outlet(0, new Atom[]{Atom.newAtom("thispatcher"), Atom.newAtom("script"), 
						Atom.newAtom("sendbox"), Atom.newAtom("noteGUI"), 
						Atom.newAtom("presentation_position"), Atom.newAtom(69), Atom.newAtom(0)});
				break;
			}
			updatedisplay(_lastDisplayed, _queueHash);
		
		}
	}
	
	//receive/update specified setting of mapping from chase/setting gui
	public void mapconfig(String setting, float val) {
		int hash = _lastHash;
		
		//the current displayed settings are not mapped but can be assigned to future mapping
		if (_lastDisplayed.theBox() == null) {
			if (setting.equals("noteoff") || setting.equals("keyup")){
				if (val < 1) 
					_lastDisplayed.sendUpOff(false);
				else
					_lastDisplayed.sendUpOff(true);
			}
			else if (setting.equals("velocity")) {
				if (val < 1)
					_lastDisplayed.sendVelocity(false);
				else
					_lastDisplayed.sendVelocity(true);
			}
			else if (setting.equals("min")) {
				_lastDisplayed.min(val);
			}
			else if (setting.equals("max")) {
				_lastDisplayed.max(val);
			}
			else if (setting.equals("precision")) {
				_lastDisplayed.precision(val);
			}
			else if (setting.equals("curve")) {
				_lastDisplayed.curve(val);
			}
			else if (setting.equals("banger")) {
				if (val < 1)
					_lastDisplayed.banger(false);
				else
					_lastDisplayed.banger(true);
			}
			else if (setting.equals("encoder")) {
				try {
					((CcMapControl)(_lastDisplayed)).encoder((int)val);
				} catch (Exception e) {}
			}
			return;
		}
		if (_lastDisplayed != null) {
			hash = _invMapping.get(_lastDisplayed.name());
		}
		if (_mapping.containsKey(hash)) {
			MapControl control = _mapping.get(hash);
			if (setting.equals("noteoff") || setting.equals("keyup")){
				if (val < 1) 
					control.sendUpOff(false);
				else
					control.sendUpOff(true);
			}
			else if (setting.equals("velocity")) {
				if (val < 1)
					control.sendVelocity(false);
				else
					control.sendVelocity(true);
			}
			else if (setting.equals("min")) {
				control.min(val);
			}
			else if (setting.equals("max")) {
				control.max(val);
			}
			else if (setting.equals("precision")) {
				control.precision(val);
			}
			else if (setting.equals("curve")) {
				control.curve(val);
			}
			else if (setting.equals("banger")) {
				if (val < 1)
					control.banger(false);
				else
					control.banger(true);
			}
			else if (setting.equals("encoder")) {
				try {
					((CcMapControl)(_lastDisplayed)).encoder((int)val);
				} catch (Exception e) {}
			}
			outputsetting(hash);
		}
	}
	
	//output (store) all mappings/settings to coll
	public void outputsettings() {
		Collection<MapControl> allsettings = _mapping.values();
		Iterator<MapControl> settings = allsettings.iterator();
		while(settings.hasNext()) {
			MapControl next = settings.next();
			if (_invMapping.containsKey(next.name())) {
				if (next.type() == MapControl.CC) {
					try {
					outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom(_invMapping.get(next.name())), 
							Atom.newAtom(_invMapping.get(next.name())), Atom.newAtom(next.name()),
							Atom.newAtom(next.type()), Atom.newAtom(next.min()), Atom.newAtom(next.max()), 
							Atom.newAtom(next.precision()), Atom.newAtom(next.curve()), 
							Atom.newAtom(((CcMapControl)next).encoder()),
							Atom.newAtom(next.sendUpOff()), Atom.newAtom(next.sendVelocity()), 
							Atom.newAtom(next.banger())});
					} catch (Exception e) {
						outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom(_invMapping.get(next.name())), 
								Atom.newAtom(_invMapping.get(next.name())), Atom.newAtom(next.name()),
								Atom.newAtom(next.type()), Atom.newAtom(next.min()), Atom.newAtom(next.max()), 
								Atom.newAtom(next.precision()), Atom.newAtom(next.curve()), 
								Atom.newAtom(next.sendUpOff()), 
								Atom.newAtom(next.sendVelocity()), Atom.newAtom(next.banger())});
					}
				}
				else {
					outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom(_invMapping.get(next.name())), 
						Atom.newAtom(_invMapping.get(next.name())), Atom.newAtom(next.name()),
						Atom.newAtom(next.type()), Atom.newAtom(next.min()), Atom.newAtom(next.max()), 
						Atom.newAtom(next.precision()), Atom.newAtom(next.curve()), Atom.newAtom(next.sendUpOff()), 
						Atom.newAtom(next.sendVelocity()), Atom.newAtom(next.banger())});
				}
			}
		}
		//output devices;
		Iterator<Character> iter = _devices.iterator();
		String devs = "";
		while (iter.hasNext()) {
			char nextchar = iter.next();
			devs = devs + nextchar;
		}
		outlet(1, new Atom[]{Atom.newAtom("store"), Atom.newAtom("mididevices"), Atom.newAtom(devs)});
	}
	
	//edit specified setting of midi/key hash code
	public void edit(int hash, String setting, float val) {
		if (_mapping.containsKey(hash)) {
			MapControl control = _mapping.get(hash);
			if (setting.equals("noteoff") || setting.equals("keyup") || setting.equals("sendmin")){
				if (val < 1) 
					control.sendUpOff(false);
				else
					control.sendUpOff(true);
			}
			else if (setting.equals("velocity")) {
				if (val < 1)
					control.sendVelocity(false);
				else
					control.sendVelocity(true);
			}
			else if (setting.equals("min")) {
				control.min(val);
			}
			else if (setting.equals("max")) {
				control.max(val);
			}
			else if (setting.equals("precision")) {
				control.precision(val);
			}
			else if (setting.equals("curve")) {
				control.curve(val);
			}
			else if (setting.equals("banger")) {
				if (val < 1)
					control.banger(false);
				else
					control.banger(true);
			}
			else if (setting.equals("encoder")) {
				try {
					((CcMapControl)(_lastDisplayed)).encoder((int)val);
				} catch (Exception e) {}
			}
			outputsetting(hash);
		}
	}
	
	//load midi device settings
	public void loadsetting(String symbol, String midiabbv) {
		enabledevices(midiabbv);
		outlet(0, new Atom[]{Atom.newAtom("mididevices"), Atom.newAtom("set"), Atom.newAtom(midiabbv)});
	}
	
	//load settings for one mapping from coll with precision, curve
	public void loadsetting(int hash, String name, int type, float min, float max, float precision, float curve,
			int encoder, boolean sendupoff, boolean sendvelocity, boolean banger) {
		MaxBox thebox = _myParent.getNamedBox(name);
		if (thebox != null){
			MapControl setting;
			if (type == MapControl.KEY) {
				setting = new KeyMapControl(thebox, name, type, min, max, precision, curve, sendupoff, sendvelocity, banger);
			}
			else if (type == MapControl.CC)
				setting = new CcMapControl(thebox, name, type, min, max, precision, curve, encoder, sendupoff, sendvelocity, banger);
			else
				setting = new MapControl(thebox, name, type, min, max, precision, curve, sendupoff, sendvelocity, banger);
			if (_invMapping.containsKey(name)) {
				_mapping.remove(_invMapping.remove(name));
			}
			if (_mapping.containsKey(hash)) {
				_invMapping.remove(_mapping.remove(hash).name());
			}
			_mapping.put(hash, setting);
			_invMapping.put(name, hash);
		}
		else {
			post("error loading settings for box named: "+name);
			//remove from coll
			outlet(1, new Atom[]{Atom.newAtom("remove"), Atom.newAtom(hash)});
		}
	}
	
	public void loadsetting(int hash, String name, int type, float min, float max, float precision, 
			int encoder, boolean sendupoff, boolean sendvelocity, boolean banger) {
		loadsetting(hash, name, type, min, max, precision, 0, encoder, sendupoff, sendvelocity, banger);
	}
	
	//load settings for one mapping from coll with precision
	public void loadsetting(int hash, String name, int type, float min, float max, float precision, 
			boolean sendupoff, boolean sendvelocity, boolean banger) {
		loadsetting(hash, name, type, min, max, precision, -1, sendupoff, sendvelocity, banger);
	}
	
	//load settings for one mapping from coll without precision
	public void loadsetting(int hash, String name, int type, float min, float max, 
			boolean sendupoff, boolean sendvelocity, boolean banger) {
		loadsetting(hash, name, type, min, max, 0, -1, sendupoff, sendvelocity, banger);
	}
	
	
	
	//remove all mappings
	public void removeall() {
		_mapping.clear();
		_invMapping.clear();
		_mapMarkers.RemoveAllBoxes();
		queuedisplay(-1);
		outlet(1, "clear");
	}
	
	//load mapping file into coll
	public void readfile() {
		outlet(1, "read");
		checkmap();
	}
	
	//output current mappings to coll and save
	public void savefile() {
		checkmap();
		outlet(1, "clear");
		outputsettings();
		outlet(1, "write");
	}
}
